#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.settings_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     20 Aug 2021, (2:30 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import tornado.log

from unmanic import config
from unmanic.libs.installation_link import Links
from unmanic.libs.library import Library
from unmanic.libs.uiserver import UnmanicDataQueues
from unmanic.libs.worker_group import WorkerGroup
from unmanic.webserver.api_v2.base_api_handler import BaseApiError, BaseApiHandler
from unmanic.webserver.api_v2.schema.schemas import RequestDatabaseItemByIdSchema, RequestLibraryByIdSchema, \
    RequestRemoteInstallationLinkConfigSchema, SettingsLibrariesListSchema, SettingsLibraryConfigReadAndWriteSchema, \
    SettingsLibraryPluginConfigExportSchema, \
    SettingsLibraryPluginConfigImportSchema, SettingsReadAndWriteSchema, \
    SettingsRemoteInstallationDataSchema, \
    SettingsRemoteInstallationLinkConfigSchema, SettingsSystemConfigSchema, \
    RequestSettingsRemoteInstallationAddressValidationSchema, SettingsWorkerGroupConfigSchema, WorkerGroupsListSchema
from unmanic.webserver.helpers import plugins


class ApiSettingsHandler(BaseApiHandler):
    config = None
    params = None
    unmanic_data_queues = None

    routes = [
        {
            "path_pattern":      r"/settings/read",
            "supported_methods": ["GET"],
            "call_method":       "get_all_settings",
        },
        {
            "path_pattern":      r"/settings/write",
            "supported_methods": ["POST"],
            "call_method":       "write_settings",
        },
        {
            "path_pattern":      r"/settings/configuration",
            "supported_methods": ["GET"],
            "call_method":       "get_system_configuration",
        },
        {
            "path_pattern":      r"/settings/link/validate",
            "supported_methods": ["POST"],
            "call_method":       "validate_remote_installation",
        },
        {
            "path_pattern":      r"/settings/worker_groups",
            "supported_methods": ["GET"],
            "call_method":       "get_all_worker_groups",
        },
        {
            "path_pattern":      r"/settings/worker_group/read",
            "supported_methods": ["POST"],
            "call_method":       "read_worker_group_config",
        },
        {
            "path_pattern":      r"/settings/worker_group/write",
            "supported_methods": ["POST"],
            "call_method":       "write_worker_group_config",
        },
        {
            "path_pattern":      r"/settings/worker_group/remove",
            "supported_methods": ["DELETE"],
            "call_method":       "remove_worker_group",
        },
        {
            "path_pattern":      r"/settings/link/read",
            "supported_methods": ["POST"],
            "call_method":       "read_link_config",
        },
        {
            "path_pattern":      r"/settings/link/write",
            "supported_methods": ["POST"],
            "call_method":       "write_link_config",
        },
        {
            "path_pattern":      r"/settings/link/remove",
            "supported_methods": ["DELETE"],
            "call_method":       "remove_link_config",
        },
        {
            "path_pattern":      r"/settings/libraries",
            "supported_methods": ["GET"],
            "call_method":       "get_all_libraries",
        },
        {
            "path_pattern":      r"/settings/library/read",
            "supported_methods": ["POST"],
            "call_method":       "read_library_config",
        },
        {
            "path_pattern":      r"/settings/library/write",
            "supported_methods": ["POST"],
            "call_method":       "write_library_config",
        },
        {
            "path_pattern":      r"/settings/library/remove",
            "supported_methods": ["DELETE"],
            "call_method":       "remove_library",
        },
        {
            "path_pattern":      r"/settings/library/export",
            "supported_methods": ["POST"],
            "call_method":       "export_library_plugin_config",
        },
        {
            "path_pattern":      r"/settings/library/import",
            "supported_methods": ["POST"],
            "call_method":       "import_library_plugin_config",
        },
    ]

    def initialize(self, **kwargs):
        self.params = kwargs.get("params")
        udq = UnmanicDataQueues()
        self.unmanic_data_queues = udq.get_unmanic_data_queues()
        self.config = config.Config()

    def get_all_settings(self):
        """
        Settings - read
        ---
        description: Returns the application settings.
        responses:
            200:
                description: 'Sample response: Returns the application settings.'
                content:
                    application/json:
                        schema:
                            SettingsReadAndWriteSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            settings = self.config.get_config_as_dict()
            response = self.build_response(
                SettingsReadAndWriteSchema(),
                {
                    "settings": settings,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def write_settings(self):
        """
        Settings - save a dictionary of settings
        ---
        description: Save a given dictionary of settings.
        requestBody:
            description: Requested a dictionary of settings to save.
            required: True
            content:
                application/json:
                    schema:
                        SettingsReadAndWriteSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(SettingsReadAndWriteSchema())

            # Get settings dict from request
            settings_dict = json_request.get('settings', {})

            # Remove config items that should not be saved through this API endpoint
            remove_settings = [
                'remote_installations'
            ]
            for remove_setting in remove_settings:
                if settings_dict.get(remove_setting):
                    del settings_dict[remove_setting]

            # Save settings - writing to file.
            # Throws exception if settings fail to save
            self.config.set_bulk_config_items(json_request.get('settings', {}))

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_system_configuration(self):
        """
        Settings - read the system configuration
        ---
        description: Returns the system configuration.
        responses:
            200:
                description: 'Sample response: Returns the system configuration.'
                content:
                    application/json:
                        schema:
                            SettingsSystemConfigSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            from unmanic.libs.system import System
            system = System()
            system_info = system.info()
            response = self.build_response(
                SettingsSystemConfigSchema(),
                {
                    "configuration": system_info,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def validate_remote_installation(self):
        """
        Settings - validate a remote installation address
        ---
        description: Validate a remote installation address
        requestBody:
            description: The details of the remote installation to validate
            required: True
            content:
                application/json:
                    schema:
                        RequestSettingsRemoteInstallationAddressValidationSchema
        responses:
            200:
                description: 'Sample response: Returns the remote installation data.'
                content:
                    application/json:
                        schema:
                            SettingsRemoteInstallationDataSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestSettingsRemoteInstallationAddressValidationSchema())

            # Fetch all data from the remote installation
            # Throws exception if the provided address is invalid
            links = Links()
            data = links.validate_remote_installation(json_request.get('address'),
                                                      auth=json_request.get('auth'),
                                                      username=json_request.get('username'),
                                                      password=json_request.get('password'))

            response = self.build_response(
                SettingsRemoteInstallationDataSchema(),
                {
                    "installation": data,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_all_worker_groups(self):
        """
        Settings - get list of all worker groups
        ---
        description: Returns a list of all worker groups.
        responses:
            200:
                description: 'Sample response: Returns a list of all worker groups.'
                content:
                    application/json:
                        schema:
                            WorkerGroupsListSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            worker_groups = WorkerGroup.get_all_worker_groups()
            response = self.build_response(
                WorkerGroupsListSchema(),
                {
                    "worker_groups": worker_groups,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def read_worker_group_config(self):
        """
        Settings - read the configuration of a worker group
        ---
        description: Read the configuration of a worker group
        requestBody:
            description: The ID of the worker group
            required: True
            content:
                application/json:
                    schema:
                        RequestDatabaseItemByIdSchema
        responses:
            200:
                description: 'Sample response: Returns the worker group configuration.'
                content:
                    application/json:
                        schema:
                            SettingsWorkerGroupConfigSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestDatabaseItemByIdSchema())

            # Fetch all data for this worker group
            worker_group = WorkerGroup(json_request.get('id'))
            if not worker_group:
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Unable to find worker group config by its ID")
                self.write_error()
                return

            response = self.build_response(
                SettingsWorkerGroupConfigSchema(),
                {
                    "id":                     worker_group.get_id(),
                    "locked":                 worker_group.get_locked(),
                    "name":                   worker_group.get_name(),
                    "number_of_workers":      worker_group.get_number_of_workers(),
                    "worker_event_schedules": worker_group.get_worker_event_schedules(),
                    "tags":                   worker_group.get_tags(),
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def write_worker_group_config(self):
        """
        Settings - write the configuration of a worker group
        ---
        description: Write the configuration of a worker group
        requestBody:
            description: The config of a worker group that is to be saved
            required: True
            content:
                application/json:
                    schema:
                        SettingsWorkerGroupConfigSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(SettingsWorkerGroupConfigSchema())

            # Write config for this worker group
            from unmanic.webserver.helpers import settings
            settings.save_worker_group_config(json_request)

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def remove_worker_group(self):
        """
        Settings - remove a worker group
        ---
        description: Remove a worker group
        requestBody:
            description: Requested a worker group to remove.
            required: True
            content:
                application/json:
                    schema:
                        RequestDatabaseItemByIdSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestDatabaseItemByIdSchema())

            # Fetch existing worker group by ID
            worker_group = WorkerGroup(json_request.get('id'))

            # Delete the worker group
            if not worker_group.delete():
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to remove worker group by its ID")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def read_link_config(self):
        """
        Settings - read the configuration of a remote installation link
        ---
        description: Read the configuration of a remote installation link
        requestBody:
            description: The UUID of the remote installation
            required: True
            content:
                application/json:
                    schema:
                        RequestRemoteInstallationLinkConfigSchema
        responses:
            200:
                description: 'Sample response: Returns the remote installation link configuration.'
                content:
                    application/json:
                        schema:
                            SettingsRemoteInstallationLinkConfigSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestRemoteInstallationLinkConfigSchema())

            # Fetch all data from the remote installation
            # Throws exception if the provided address is invalid
            links = Links()
            data = links.read_remote_installation_link_config(json_request.get('uuid'))

            response = self.build_response(
                SettingsRemoteInstallationLinkConfigSchema(),
                {
                    "link_config":                     {
                        "address":                         data.get('address'),
                        "auth":                            data.get('auth'),
                        "username":                        data.get('username'),
                        "password":                        data.get('password'),
                        "available":                       data.get('available', False),
                        "name":                            data.get('name'),
                        "version":                         data.get('version'),
                        "last_updated":                    data.get('last_updated', 1),
                        "enable_receiving_tasks":          data.get('enable_receiving_tasks'),
                        "enable_sending_tasks":            data.get('enable_sending_tasks'),
                        "enable_task_preloading":          data.get('enable_task_preloading'),
                        "preloading_count":                data.get('preloading_count'),
                        "enable_checksum_validation":      data.get('enable_checksum_validation'),
                        "enable_config_missing_libraries": data.get('enable_config_missing_libraries'),
                        "enable_distributed_worker_count": data.get('enable_distributed_worker_count', False),
                    },
                    "distributed_worker_count_target": data.get('distributed_worker_count_target', 0),
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def write_link_config(self):
        """
        Settings - write the configuration of a remote installation link
        ---
        description: Write the configuration of a remote installation link
        requestBody:
            description: The UUID of the remote installation and its configuration
            required: True
            content:
                application/json:
                    schema:
                        SettingsRemoteInstallationLinkConfigSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(SettingsRemoteInstallationLinkConfigSchema())

            # Update a single remote installation config by matching the UUID
            links = Links()
            links.update_single_remote_installation_link_config(json_request.get('link_config'),
                                                                json_request.get('distributed_worker_count_target', 0))

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def remove_link_config(self):
        """
        Settings - remove a configuration for a remote installation link
        ---
        description: Remove a configuration for a remote installation link
        requestBody:
            description: Requested a remote installation link to remove.
            required: True
            content:
                application/json:
                    schema:
                        RequestRemoteInstallationLinkConfigSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestRemoteInstallationLinkConfigSchema())

            # Delete the remote installation using the given uuid
            links = Links()
            if not links.delete_remote_installation_link_config(json_request.get('uuid')):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to remove link by its uuid")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def get_all_libraries(self):
        """
        Settings - get list of all libraries
        ---
        description: Returns a list of all libraries.
        responses:
            200:
                description: 'Sample response: Returns a list of all libraries.'
                content:
                    application/json:
                        schema:
                            SettingsLibrariesListSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            libraries = Library.get_all_libraries()
            response = self.build_response(
                SettingsLibrariesListSchema(),
                {
                    "libraries": libraries,
                }
            )
            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def read_library_config(self):
        """
        Settings - read the configuration of one library
        ---
        description: Read the configuration of one library
        requestBody:
            description: The ID of the library
            required: True
            content:
                application/json:
                    schema:
                        RequestLibraryByIdSchema
        responses:
            200:
                description: 'Sample response: Returns the remote installation link configuration.'
                content:
                    application/json:
                        schema:
                            SettingsLibraryConfigReadAndWriteSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestLibraryByIdSchema())

            library_settings = {
                "library_config": {
                    "id":                 0,
                    "name":               '',
                    "path":               '/',
                    "enable_remote_only": False,
                    "enable_scanner":     False,
                    "enable_inotify":     False,
                    "priority_score":     0,
                },
                "plugins":        {
                    "enabled_plugins": [],
                }
            }
            if json_request.get('id'):
                # Read the library
                library_config = Library(json_request.get('id'))
                library_settings = {
                    "library_config": {
                        "id":                 library_config.get_id(),
                        "name":               library_config.get_name(),
                        "path":               library_config.get_path(),
                        "locked":             library_config.get_locked(),
                        "enable_remote_only": library_config.get_enable_remote_only(),
                        "enable_scanner":     library_config.get_enable_scanner(),
                        "enable_inotify":     library_config.get_enable_inotify(),
                        "priority_score":     library_config.get_priority_score(),
                        "tags":               library_config.get_tags(),
                    },
                    "plugins":        {
                        "enabled_plugins": library_config.get_enabled_plugins(),
                    }
                }

            response = self.build_response(
                SettingsLibraryConfigReadAndWriteSchema(),
                library_settings
            )

            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def write_library_config(self):
        """
        Settings - write the configuration of one library
        ---
        description: Write the configuration of one library
        requestBody:
            description: Requested a dictionary of settings to save.
            required: True
            content:
                application/json:
                    schema:
                        SettingsLibraryConfigReadAndWriteSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(SettingsLibraryConfigReadAndWriteSchema())

            # Save settings
            from unmanic.webserver.helpers import settings
            library_config = json_request['library_config']
            plugin_config = json_request.get('plugins', {})
            library_id = library_config.get('id', 0)
            if not settings.save_library_config(library_id, library_config=library_config, plugin_config=plugin_config):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to write library config")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def remove_library(self):
        """
        Settings - remove a library
        ---
        description: Remove a library
        requestBody:
            description: Requested a library to remove.
            required: True
            content:
                application/json:
                    schema:
                        RequestLibraryByIdSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestLibraryByIdSchema())

            # Fetch existing library by ID
            library = Library(json_request.get('id'))

            # Delete the library
            if not library.delete():
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to remove library by its ID")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def export_library_plugin_config(self):
        """
        Settings - export the plugin configuration of one library
        ---
        description: Export the plugin configuration of one library
        requestBody:
            description: The ID of the library
            required: True
            content:
                application/json:
                    schema:
                        RequestLibraryByIdSchema
        responses:
            200:
                description: 'Sample response: Returns the remote installation link configuration.'
                content:
                    application/json:
                        schema:
                            SettingsLibraryPluginConfigExportSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(RequestLibraryByIdSchema())

            # Fetch library config
            library_config = Library.export(json_request.get('id'))

            response = self.build_response(
                SettingsLibraryPluginConfigExportSchema(),
                library_config
            )

            self.write_success(response)
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()

    def import_library_plugin_config(self):
        """
        Settings - import the plugin configuration of one library
        ---
        description: Import the configuration of one library
        requestBody:
            description: Requested a dictionary of settings to save.
            required: True
            content:
                application/json:
                    schema:
                        SettingsLibraryPluginConfigImportSchema
        responses:
            200:
                description: 'Successful request; Returns success status'
                content:
                    application/json:
                        schema:
                            BaseSuccessSchema
            400:
                description: Bad request; Check `messages` for any validation errors
                content:
                    application/json:
                        schema:
                            BadRequestSchema
            404:
                description: Bad request; Requested endpoint not found
                content:
                    application/json:
                        schema:
                            BadEndpointSchema
            405:
                description: Bad request; Requested method is not allowed
                content:
                    application/json:
                        schema:
                            BadMethodSchema
            500:
                description: Internal error; Check `error` for exception
                content:
                    application/json:
                        schema:
                            InternalErrorSchema
        """
        try:
            json_request = self.read_json_request(SettingsLibraryPluginConfigImportSchema())

            # Save settings
            from unmanic.webserver.helpers import settings
            library_config = json_request.get('library_config')
            plugin_config = json_request.get('plugins', {})
            library_id = json_request.get('library_id')
            if not settings.save_library_config(library_id, library_config=library_config, plugin_config=plugin_config):
                self.set_status(self.STATUS_ERROR_INTERNAL, reason="Failed to import library config")
                self.write_error()
                return

            self.write_success()
            return
        except BaseApiError as bae:
            tornado.log.app_log.error("BaseApiError.{}: {}".format(self.route.get('call_method'), str(bae)))
            return
        except Exception as e:
            self.set_status(self.STATUS_ERROR_INTERNAL, reason=str(e))
            self.write_error()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.history_api.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     25 Oct 2020, (8:49 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import json
import os
import time
import tornado.web
import tornado.log
import tornado.routing

from unmanic import config
from unmanic.webserver.api_v1.base_api_handler import BaseApiHandler

from unmanic.libs import history, task, common
from unmanic.webserver.helpers import completed_tasks


class ApiHistoryHandler(BaseApiHandler):
    name = None
    config = None
    params = None

    routes = [
        {
            "supported_methods": ["GET", "POST"],
            "call_method":       "fetch_by_id",
            "path_pattern":      r"/api/v1/history/id/(?P<id>[0-9]+)?",
        },
        {
            "supported_methods": ["POST"],
            "call_method":       "manage_historic_tasks_list",
            "path_pattern":      r"/api/v1/history/list",
        },
    ]

    def initialize(self, **kwargs):
        self.name = 'history_api'
        self.config = config.Config()
        self.params = kwargs.get("params")

    def set_default_headers(self):
        """Set the default response header to be JSON."""
        self.set_header("Content-Type", 'application/json; charset="utf-8"')

    def post(self, path):
        self.action_route()

    def fetch_by_id(self, *args, **kwargs):
        # TODO: add ability to fetch by id
        pass

    def manage_historic_tasks_list(self, *args, **kwargs):
        request_dict = json.loads(self.request.body)

        # Delete a list of historical tasks.
        #   (on success will continue to return the current list of historical tasks)
        if request_dict.get("customActionName") == "delete-from-history":
            success = self.delete_historic_tasks(request_dict.get("id"))
            if not success:
                self.write(json.dumps({"success": False}))
                return

        # Return a list of historical tasks based on the request JSON body
        results = self.prepare_filtered_historic_tasks(request_dict)
        self.write(json.dumps(results))

    def delete_historic_tasks(self, historic_task_ids):
        """
        Deletes a list of historic tasks

        :param historic_task_ids:
        :return:
        """
        # Fetch historical tasks
        history_logging = history.History()
        # Delete by ID
        return history_logging.delete_historic_tasks_recursively(id_list=historic_task_ids)

    def prepare_filtered_historic_tasks(self, request_dict):
        """
        Returns a object of historical records filtered and sorted
        according to the provided request.

        :param request_dict:
        :return:
        """

        # Generate filters for query
        draw = request_dict.get('draw')
        start = request_dict.get('start')
        length = request_dict.get('length')

        search = request_dict.get('search')
        search_value = search.get("value")

        # Get sort order
        filter_order = request_dict.get('order')[0]
        order_direction = filter_order.get('dir')
        columns = request_dict.get('columns')
        order_column_name = columns[filter_order.get('column')].get('name')
        order = {
            "column": order_column_name,
            "dir":    order_direction,
        }

        # Fetch historical tasks
        history_logging = history.History()
        # Get total count
        records_total_count = history_logging.get_total_historic_task_list_count()
        # Get quantity after filters (without pagination)
        records_filtered_count = history_logging.get_historic_task_list_filtered_and_sorted(order=order, start=0, length=0,
                                                                                            search_value=search_value).count()
        # Get filtered/sorted results
        task_results = history_logging.get_historic_task_list_filtered_and_sorted(order=order, start=start, length=length,
                                                                                  search_value=search_value)

        # Build return data
        return_data = {
            "draw":            draw,
            "recordsTotal":    records_total_count,
            "recordsFiltered": records_filtered_count,
            "successCount":    0,
            "failedCount":     0,
            "data":            []
        }

        # Iterate over historical tasks and append them to the task data
        for task in task_results:
            # Set params as required in template
            item = {
                'id':           task['id'],
                'selected':     False,
                'finish_time':  time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(task['finish_time'])),
                'task_label':   task['task_label'],
                'task_success': task['task_success'],
            }
            # Increment counters
            if item['task_success']:
                return_data["successCount"] += 1
            else:
                return_data["failedCount"] += 1
            return_data["data"].append(item)

        # Return results
        return return_data



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.eventmonitor.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     25 Feb 2021, (10:06 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import queue
import threading
import time

from unmanic import config
from unmanic.libs.library import Library
from unmanic.libs.plugins import PluginsHandler

try:
    from watchdog.observers import Observer
    from watchdog.events import FileSystemEventHandler

    event_monitor_module = 'watchdog'
except ImportError:
    class Observer(object):
        pass


    class FileSystemEventHandler(object):
        pass


    event_monitor_module = None

from unmanic.libs import common, unlogger
from unmanic.libs.filetest import FileTest


class EventHandler(FileSystemEventHandler):
    """
    Handle any library file modification events

    This watchdog library is not strictly inline with inotify...
    It does not watch MOVED_TO, CREATED, DELETE, MODIFIED, CLOSE_WRITE like the previous library did.
    Instead the outputs for a Unix FS are as follows:
        - Read = []                                             : No events are triggered for reading a file
        - Modify = ["modified", "closed"]                       :
        - Move (atomic) = ["created", "modified"]               :
        - Move (non-atomic) = ["created", "modified", "closed"] : Lots of 'modified' events as the file was "copied"
        - Copy = ["created", "modified", "closed"]              : ^ ditto
        - Create = ["created", "modified", "closed"]            : ^ ditto
        - Delete = ["deleted", "modified"]                      :
        - Hardlink = ["created", "modified"]                    :

    From this, the only event we really need to monitor is the "created" and "closed" events.
    """

    def __init__(self, files_to_test, library_id):
        self.name = "EventProcessor"
        self.files_to_test = files_to_test
        self.library_id = library_id
        self.logger = None
        self.abort_flag = threading.Event()
        self.abort_flag.clear()

    def _log(self, message, message2='', level="info"):
        if not self.logger:
            unmanic_logging = unlogger.UnmanicLogger.__call__()
            self.logger = unmanic_logging.get_logger(self.name)
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def on_any_event(self, event):
        if event.event_type in ["created", "closed"]:
            # Ensure event was not for a directory
            if event.is_directory:
                self._log("Detected event is for a directory. Ignoring...", level="debug")
            else:
                self._log("Detected '{}' event on file path '{}'".format(event.event_type, event.src_path))
                self.files_to_test.put({
                    'src_path':   event.src_path,
                    'library_id': self.library_id,
                })


class EventMonitorManager(threading.Thread):
    """
    EventMonitorManager

    Manage the EventProcessor thread.
    If the settings for enabling the EventProcessor changes, this manager
    class will stop or start the EventProcessor thread accordingly.

    """

    def __init__(self, data_queues, event):
        super(EventMonitorManager, self).__init__(name='EventMonitorManager')
        self.name = "EventMonitorManager"
        self.data_queues = data_queues
        self.settings = config.Config()
        self.logger = None
        self.event = event

        # Create an event queue
        self.files_to_test = queue.Queue()

        self.abort_flag = threading.Event()
        self.abort_flag.clear()

        self.event_observer_thread = None
        self.event_observer_threads = []

    def _log(self, message, message2='', level="info"):
        if not self.logger:
            unmanic_logging = unlogger.UnmanicLogger.__call__()
            self.logger = unmanic_logging.get_logger(self.name)
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        self.abort_flag.set()

    def run(self):
        self._log("Starting EventMonitorManager loop")
        while not self.abort_flag.is_set():
            self.event.wait(.5)

            if not self.system_configuration_is_valid():
                self.event.wait(2)
                continue

            if not self.files_to_test.empty():
                item = self.files_to_test.get()
                pathname = item.get('src_path')
                library_id = item.get('library_id')
                self.manage_event_queue(pathname, library_id)
                continue

            # Check if monitor is enabled for at least one library
            enable_inotify = False
            for lib_info in Library.get_all_libraries():
                try:
                    library = Library(lib_info['id'])
                except Exception as e:
                    self._log("Unable to fetch library config for ID {}".format(lib_info['id']), level='exception')
                    continue
                # Check if the library is configured for remote files only
                if library.get_enable_remote_only():
                    # This library is configured to receive remote files only... Never enable the file monitor
                    continue
                # Check if file monitor is enabled on any library
                if library.get_enable_inotify():
                    enable_inotify = True

            # If at least library has the monitor enabled, then start it. Otherwise stop the monitor process
            if enable_inotify:
                # If enabled, ensure it is running and start it if it is not
                if not self.event_observer_thread:
                    self.start_event_processor()
            else:
                # If not enabled, ensure the EventProcessor is not running and stop it if it is
                if self.event_observer_thread:
                    self.stop_event_processor()
            # Add delay
            self.event.wait(2)

        self.stop_event_processor()
        self._log("Leaving EventMonitorManager loop...")

    def system_configuration_is_valid(self):
        """
        Check and ensure the system configuration is correct for running

        :return:
        """
        valid = True
        plugin_handler = PluginsHandler()
        if plugin_handler.get_incompatible_enabled_plugins(self.data_queues.get('frontend_messages')):
            valid = False
        if not Library.within_library_count_limits(self.data_queues.get('frontend_messages')):
            valid = False
        return valid

    def start_event_processor(self):
        """
        Start the EventProcessor thread if it is not already running.

        :return:
        """
        if not self.event_observer_thread:
            monitoring_path = False
            self.event_observer_thread = Observer()
            for lib_info in Library.get_all_libraries():
                self.event.wait(.2)
                try:
                    library = Library(lib_info['id'])
                except Exception as e:
                    self._log("Unable to fetch library config for ID {}".format(lib_info['id']), level='exception')
                    continue
                # Check if the library is configured for remote files only
                if library.get_enable_remote_only():
                    # This library is configured to receive remote files only... Never enable the file monitor
                    continue
                # Check if library scanner is enabled on any library
                if library.get_enable_inotify():
                    library_path = library.get_path()
                    if not os.path.exists(library_path):
                        continue
                    self._log("Adding library path to monitor '{}'".format(library_path))
                    event_handler = EventHandler(self.files_to_test, library.get_id())
                    self.event_observer_thread.schedule(event_handler, library_path, recursive=True)
                    monitoring_path = True
            # Only start observer if a path was added to be monitored
            if monitoring_path:
                self._log("EventMonitorManager spawning EventProcessor thread...")
                self.event_observer_thread.start()
        else:
            self._log("Given signal to start the EventProcessor thread, but it is already running....")

    def stop_event_processor(self):
        """
        Stop the EventProcessor thread if it is running.

        :return:
        """
        if self.event_observer_thread:
            self._log("EventMonitorManager stopping EventProcessor thread...")

            self._log("Sending thread EventProcessor abort signal")
            self.event_observer_thread.stop()

            self._log("Waiting for thread EventProcessor to stop")
            self.event_observer_thread.join()
            self._log("Thread EventProcessor has successfully stopped")
        else:
            self._log("Given signal to stop the EventProcessor thread, but it is not running...")

        self.event_observer_thread = None

    def manage_event_queue(self, pathname, library_id):
        """
        Manage all monitored events

        Unlike the library scanner, all events are processed sequentially one at a time.
        This avoids a file being added twice on 2 events.

        :param pathname:
        :param library_id:
        :return:
        """
        # Test file to be added to task list. Add it if required
        try:
            file_test = FileTest(library_id)
            result, issues, priority_score = file_test.should_file_be_added_to_task_list(pathname)
            # Log any error messages
            for issue in issues:
                if type(issue) is dict:
                    self._log(issue.get('message'))
                else:
                    self._log(issue)
            # If file needs to be added, then add it
            if result:
                self.__add_path_to_queue(pathname, library_id, priority_score)
        except UnicodeEncodeError:
            self._log("File contains Unicode characters that cannot be processed. Ignoring.", level="warning")
        except Exception as e:
            self._log("Exception testing file path in {}. Ignoring.".format(self.name), message2=str(e), level="exception")

    def __add_path_to_queue(self, pathname, library_id, priority_score):
        """
        Add a given path to the pending task queue

        :param pathname:
        :param library_id:
        :param priority_score:
        :return:
        """
        self.data_queues.get('inotifytasks').put({
            'pathname':       pathname,
            'library_id':     library_id,
            'priority_score': priority_score,
        })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.main.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     21 Sep 2019, (2:08 PM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

from . import audio_codecs
from . import subtitle_codecs
from . import video_codecs
from .lib import cli


class Info(object):
    """
    Info

    Provide information on FFMPEG commands and configuration
    """
    available_encoders = None
    available_decoders = None

    @staticmethod
    def versions():
        """
        Return the system ffmpeg version as a string

        :return:
        """
        return cli.ffmpeg_version_info()

    def file_probe(self, vid_file_path):
        """
        Probe media file and return result dictionary

        :param vid_file_path:
        :return:
        """
        # TODO: Move this to a new "Probe" class
        return cli.ffprobe_file(vid_file_path)

    def get_available_ffmpeg_encoders(self):
        """
        Sets a dictionary of encoders supported by ffmpeg
        """
        available_audio_encoders = {}
        available_subtitle_encoders = {}
        available_video_encoders = {}

        # Get raw ffmpeg output of available encoders
        info = cli.ffmpeg_available_encoders()

        # Sort through the lines and create a dictionary of audio, subtitle and video encoders
        for line in info.splitlines():
            line = line.rstrip().lstrip()
            if line.startswith('A') and line != 'A..... = Audio':
                # Audio encoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_audio_encoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }
            elif line.startswith('S') and line != 'S..... = Subtitle':
                # Subtitle encoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_subtitle_encoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }
            elif line.startswith('V') and line != 'V..... = Video':
                # Video encoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_video_encoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }

        # Combine dictionaries into one
        self.available_encoders = {
            'audio':    available_audio_encoders,
            'subtitle': available_subtitle_encoders,
            'video':    available_video_encoders
        }

        return self.available_encoders

    def get_available_ffmpeg_decoders(self):
        """
        Sets a dictionary of decoders supported by ffmpeg
        """
        available_audio_decoders = {}
        available_subtitle_decoders = {}
        available_video_decoders = {}

        # Get raw ffmpeg output of available decoders
        info = cli.ffmpeg_available_decoders()

        # Sort through the lines and create a dictionary of audio, subtitle and video decoders
        for line in info.splitlines():
            line = line.rstrip().lstrip()
            if line.startswith('A') and line != 'A..... = Audio':
                # Audio decoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_audio_decoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }
            elif line.startswith('S') and line != 'S..... = Subtitle':
                # Subtitle decoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_subtitle_decoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }
            elif line.startswith('V') and line != 'V..... = Video':
                # Video decoder
                data = line.split()
                capabilities = data.pop(0)
                codec = data.pop(0)
                available_video_decoders[codec] = {
                    'capabilities': capabilities,
                    'description':  " ".join(data)
                }

        # Combine dictionaries into one
        self.available_decoders = {
            'audio':    available_audio_decoders,
            'subtitle': available_subtitle_decoders,
            'video':    available_video_decoders
        }

        return self.available_decoders

    def get_available_ffmpeg_hw_acceleration_methods(self):
        methods = []

        # Get raw ffmpeg output of available encoders
        info = cli.ffmpeg_available_hw_acceleration_methods()

        # Sort through the lines and create a list of methods
        for line in info.splitlines():
            line = line.rstrip().lstrip()
            if not line or line.startswith('Hardware acceleration'):
                continue
            else:
                methods.append(line)

        return methods

    def get_ffmpeg_audio_encoders(self):
        """
        Fetch all audio encoders supported by ffmpeg

        :return:
        """
        if self.available_encoders is None:
            self.get_available_ffmpeg_encoders()
        return self.available_encoders['audio']

    def get_ffmpeg_subtitle_encoders(self):
        """
        Fetch all subtitle encoders supported by ffmpeg

        :return:
        """
        if self.available_encoders is None:
            self.get_available_ffmpeg_encoders()
        return self.available_encoders['subtitle']

    def get_ffmpeg_video_encoders(self):
        """
        Fetch all video encoders supported by ffmpeg

        :return:
        """
        if self.available_encoders is None:
            self.get_available_ffmpeg_encoders()
        return self.available_encoders['video']

    def filter_available_encoders_for_codec(self, codec_encoders, codec_type):
        """
        Filter a given list of encoders. Removes any that are not available with FFMPEG

        :param codec_type:
        :param codec_encoders:
        :return:
        """
        available_encoders = {}
        if codec_type == 'audio':
            available_encoders = self.get_ffmpeg_audio_encoders()
        elif codec_type == 'subtitle':
            available_encoders = self.get_ffmpeg_subtitle_encoders()
        elif codec_type == 'video':
            available_encoders = self.get_ffmpeg_video_encoders()
        # Iterate through the list of encoders.
        for encoder in codec_encoders:
            # Check if ffmpeg has that encoder
            if encoder not in available_encoders:
                # Encoder is not available, remove it from the list
                codec_encoders.remove(encoder)
        return codec_encoders

    def get_all_supported_codecs_of_type(self, codec_type):
        """
        Fetch a list of supported codecs and
        return a dictionary of their data

        :return:
        """
        codec_dict = {}
        return_codec_dict = {}
        if codec_type == 'audio':
            codec_dict = audio_codecs.get_all_audio_codecs()
        elif codec_type == 'subtitle':
            codec_dict = audio_codecs.get_all_audio_codecs()
        elif codec_type == 'video':
            codec_dict = video_codecs.get_all_video_codecs()
        # Iterate through the list of codecs.
        for codec_name in codec_dict:
            codec = codec_dict[codec_name]
            # Get list of encoders for this codec that are available in ffmpeg
            codec_encoders = self.filter_available_encoders_for_codec(codec['encoders'], codec_type)
            # Check if any encoders were found
            if not codec_encoders:
                continue
            # At least one encoder is found for that codec.
            # Add codec to codec_list if one encoder exists
            return_codec_dict[codec_name] = codec
        return return_codec_dict

    def get_all_supported_video_codecs(self):
        """
        Fetch a list of supported video codecs and
        return a dictionary of their data

        :return:
        """
        return_codec_dict = {}
        codec_dict = video_codecs.get_all_video_codecs()
        for codec_name in codec_dict:
            codec = codec_dict[codec_name]
            # Get list of encoders for this codec that are available in ffmpeg
            codec_encoders = self.filter_available_encoders_for_codec(codec['encoders'], 'video')
            # Check if any encoders were found
            if not codec_encoders:
                continue
            # At least one encoder is found for that codec.
            # Add codec to codec_list if one encoder exists
            return_codec_dict[codec_name] = codec
        return return_codec_dict

    def get_all_supported_codecs(self):
        supported_audio_codecs = self.get_all_supported_codecs_of_type('audio')
        # TODO: Subtitles
        supported_video_codecs = self.get_all_supported_codecs_of_type('video')

        # Combine dictionaries into one and return
        return {
            'audio':    supported_audio_codecs,
            'subtitle': {},
            'video':    supported_video_codecs
        }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.subtitle_handle.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     19 Sep 2019, (5:23 PM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""


class SubtitleHandle(object):
    """
    SubtitleHandle

    Handle FFMPEG operations pertaining to subtitle streams
    """

    def __init__(self, file_probe, container):
        self.file_probe = file_probe
        self.container = container
        self.subtitle_args = {}

        # Configurable settings
        self.remove_subtitle_streams = False

        # Check if destination container supports subtitles
        if not container.container_supports_subtitles():
            # Destination container does not support subtitles,
            # Force them to be removed
            self.remove_subtitle_streams = True

    def args(self):
        """
        Return a dictionary of streams to map and streams to encode
        :return:
        """
        # Read stream data
        self.subtitle_args['streams_to_map'] = []
        self.subtitle_args['streams_to_encode'] = []
        subtitle_tracks_count = 0
        for stream in self.file_probe['streams']:
            # If this is a subtitle stream, then process the args
            if stream['codec_type'] == 'subtitle':
                # Remove subtitles means add no args
                if self.remove_subtitle_streams:
                    continue

                # Add stream
                # Check container for support of current stream (If copy is possible)
                # TODO: Add support for user selection of subtitle format
                supported_subtitles = self.container.supported_subtitles()
                # TODO: Select best/or configured subtitle codec, then fetch that codec class.
                #       Use the subtitle class rather than this array
                if stream['codec_name'] in supported_subtitles:
                    # If dest container supports the current subtitle codec, just copy it
                    self.subtitle_args['streams_to_encode'] = self.subtitle_args['streams_to_encode'] + [
                        "-c:s:{}".format(subtitle_tracks_count), "copy"
                    ]
                    subtitle_tracks_count += 1
                else:
                    # The dest container does not support the current subtitle stream.
                    # Transcode the stream to a format that the destination container does support
                    # TODO: Check if it can be re-encoded. It is not possible to switch between image and text format
                    # If dest container supports the current subtitle codec, just copy it
                    # unsupported subtitles will need to be removed, otherwise ffmpeg will not convert
                    unsupported_subtitles = self.container.unsupported_subtitles()
                    if stream['codec_name'] in unsupported_subtitles:
                        continue
                    else:
                        self.subtitle_args['streams_to_encode'] = self.subtitle_args['streams_to_encode'] + [
                            "-c:s:{}".format(subtitle_tracks_count), "{}".format(supported_subtitles[0])
                        ]
                        subtitle_tracks_count += 1

                # Map this stream if it was marked above as compatible with the destination
                self.subtitle_args['streams_to_map'] = self.subtitle_args['streams_to_map'] + [
                    "-map", "0:{}".format(stream['index'])
                ]

        return self.subtitle_args

    def remove_subtitles(self):
        """
        Remove the subtitles stream from result file

        :return:
        """
        self.remove_subtitle_streams = True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.completed_tasks.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     24 Jul 2021, (9:34 AM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import time
from datetime import date, datetime

from unmanic.libs import common, history, task


def prepare_filtered_completed_tasks(params):
    """
    Returns a object of historical records filtered and sorted
    according to the provided request.

    :param params:
    :return:
    """
    start = params.get('start', 0)
    length = params.get('length', 0)

    search_value = params.get('search_value', '')
    status = params.get('status', 'all')

    order = params.get('order', {
        "column": 'finish_time',
        "dir":    'desc',
    })

    # Define filters
    task_success = None
    if status == 'success':
        task_success = True
    elif status == 'failed':
        task_success = False

    after_time = None
    if params.get('after'):
        after_time = datetime.strptime(params.get('after'), '%Y-%m-%dT%H:%M:%S').timestamp()

    before_time = None
    if params.get('before'):
        before_time = datetime.strptime(params.get('before'), '%Y-%m-%dT%H:%M:%S').timestamp()

    # Fetch historical tasks
    history_logging = history.History()
    # Get total count
    records_total_count = history_logging.get_total_historic_task_list_count()
    # Get total success count
    records_total_success_count = history_logging.get_historic_task_list_filtered_and_sorted(task_success=True).count()
    # Get total failed count
    records_total_failed_count = history_logging.get_historic_task_list_filtered_and_sorted(task_success=False).count()
    # Get quantity after filters (without pagination)
    records_filtered_count = history_logging.get_historic_task_list_filtered_and_sorted(order=order, start=0, length=0,
                                                                                        search_value=search_value,
                                                                                        task_success=task_success,
                                                                                        after_time=after_time,
                                                                                        before_time=before_time).count()
    # Get filtered/sorted results
    task_results = history_logging.get_historic_task_list_filtered_and_sorted(order=order, start=start, length=length,
                                                                              search_value=search_value,
                                                                              task_success=task_success, after_time=after_time,
                                                                              before_time=before_time)

    # Build return data
    return_data = {
        "recordsTotal":    records_total_count,
        "recordsFiltered": records_filtered_count,
        "successCount":    records_total_success_count,
        "failedCount":     records_total_failed_count,
        "results":         []
    }

    # Iterate over tasks and append them to the task data
    for task in task_results:
        # Set params as required in template
        item = {
            'id':           task['id'],
            'task_label':   task['task_label'],
            'task_success': task['task_success'],
            'finish_time':  task['finish_time'],
        }
        return_data["results"].append(item)

    # Return results
    return return_data


def remove_completed_tasks(completed_task_ids):
    """
    Removes a list of completed tasks

    :param completed_task_ids:
    :return:
    """
    # Delete by ID
    task_handler = history.History()
    return task_handler.delete_historic_tasks_recursively(id_list=completed_task_ids)


def add_historic_tasks_to_pending_tasks_list(historic_task_ids, library_id=None):
    """
    Adds a list of historical tasks to the pending tasks list.

    :param historic_task_ids:
    :param library_id:
    :return:
    """
    errors = {}
    # Fetch historical tasks
    history_logging = history.History()
    # Get total count
    records_by_id = history_logging.get_current_path_of_historic_tasks_by_id(id_list=historic_task_ids)
    for record in records_by_id:
        record_errors = []
        # Fetch the abspath name
        abspath = os.path.abspath(record.get("abspath"))

        # Ensure path exists
        if not os.path.exists(abspath):
            errors[record.get("id")] = "Path does not exist - '{}'".format(abspath)
            continue

        # Create a new task
        new_task = task.Task()

        if not new_task.create_task_by_absolute_path(abspath, library_id=library_id):
            # If file exists in task queue already this will return false.
            # Do not carry on.
            errors[record.get("id")] = "File already in task queue - '{}'".format(abspath)

        continue
    return errors


def read_command_log_for_task(task_id):
    data = {
        'command_log':       '',
        'command_log_lines': [],
    }
    task_handler = history.History()
    task_data = task_handler.get_historic_task_data_dictionary(task_id=task_id)
    if not task_data:
        return data

    for command_log in task_data.get('completedtaskscommandlogs_set', []):
        data['command_log'] += command_log['dump']
        data['command_log_lines'] += format_ffmpeg_log_text(command_log['dump'].split("\n"))

    return data


def format_ffmpeg_log_text(log_lines):
    return_list = []
    pre_text = False
    headers = ['RUNNER:', 'COMMAND:', 'LOG:', 'WORKER TERMINATED!', 'PLUGIN FAILED!', 'REMOTE TASK FAILED!',
               'REMOTE LINK MANAGER TERMINATED!']
    for i, line in enumerate(log_lines):
        line_text = line

        # Add PRE to lines
        if line_text and pre_text and line_text.rstrip() not in headers:
            line_text = '<pre>{}</pre>'.format(line_text)

        # Add bold to headers
        if line_text.rstrip() not in headers:
            line_text = line_text
        else:
            if line_text.rstrip() in ['WORKER TERMINATED!', 'PLUGIN FAILED!', 'REMOTE TASK FAILED!',
                                      'REMOTE LINK MANAGER TERMINATED!']:
                line_text = '<b><span class="terminated">{}</span></b>'.format(line_text)
            else:
                line_text = '<b>{}</b>'.format(line_text)

        # Replace leading whitespace
        stripped = line.lstrip()
        line_text = "&nbsp;" * (len(line) - len(stripped)) + line_text

        # If log section is COMMAND:
        if 'RUNNER:' in line_text:
            # prepend a horizontal rule
            return_list.append("<hr>")
            pre_text = False
        elif 'COMMAND:' in line_text:
            pre_text = True
        elif 'LOG:' in line_text:
            pre_text = False
        return_list.append(line_text)
    return return_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.filebrowser.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     13 Aug 2021, (2:44 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import string

from unmanic.libs import common


def fetch_windows_drives():
    # Credit: https://stackoverflow.com/a/37761506
    return ['%s:' % d for d in string.ascii_uppercase if os.path.exists('%s:' % d)]


class DirectoryListing(object):
    """
    DirectoryListing

    Handle directory listing on the host running Unmanic
    """

    def __init__(self, list_type=None):
        self.list_type = 'all'
        if list_type:
            self.list_type = list_type

    def fetch_path_data(self, path):
        """
        Returns an object filled with data pertaining to a particular path

        :param path:
        :param list_type:
        :return:
        """
        directories = []
        files = []
        if self.list_type == "directories" or self.list_type == "all":
            directories = self.fetch_directories(path)
        if self.list_type == "files" or self.list_type == "all":
            files = self.fetch_files(path)
        path_data = {
            "current_path": path,
            "list_type":    self.list_type,
            "directories":  directories,
            "files":        files,
        }
        return path_data

    @staticmethod
    def fetch_directories(path):
        """
        Fetch a list of directory objects based on a given path

        :param path:
        :return:
        """
        results = []
        if os.path.exists(path):
            # check if this is a root path or if it has a parent
            parent_path = os.path.join(path, '..')
            if os.path.exists(parent_path) and os.path.abspath(parent_path) != path:
                # Path has a parent, Add the double dots
                results.append(
                    {
                        "name":      "..",
                        "full_path": os.path.abspath(parent_path),
                    }
                )
            elif os.name == "nt":
                # Windows allow selection of drives as parent to root directory
                results.append(
                    {
                        "name":      "..",
                        "full_path": "",
                    }
                )
            try:
                for item in sorted(os.listdir(path)):
                    abspath = os.path.abspath(os.path.join(path, item))
                    if os.path.isdir(abspath):
                        results.append(
                            {
                                "name":      item,
                                "full_path": abspath,
                            }
                        )
            except PermissionError:
                pass
        elif os.name == "nt":
            # If path does not exist and OS is Windows, then list the available drives
            for drive in fetch_windows_drives():
                results.append(
                    {
                        "name":      drive,
                        "full_path": os.path.abspath(os.path.join(drive, os.sep)),
                    }
                )
        else:
            # Path doesn't exist!
            # Just return the root dir as the first directory option
            root_path = common.get_default_root_path()
            results.append(
                {
                    "name":      root_path,
                    "full_path": root_path,
                }
            )
        return results

    @staticmethod
    def fetch_files(path):
        """
        Fetch a list of file objects based on a given path

        :param path:
        :return:
        """
        results = []
        if os.path.exists(path):
            for item in sorted(os.listdir(path)):
                abspath = os.path.abspath(os.path.join(path, item))
                if os.path.isfile(abspath):
                    results.append(
                        {
                            "name":      item,
                            "full_path": abspath,
                        }
                    )
        return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.filetest.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     28 Mar 2021, (7:28 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import os
import queue
import threading
import time
from copy import deepcopy

from unmanic import config
from unmanic.libs import history, common, unlogger
from unmanic.libs.plugins import PluginsHandler


class FileTest(object):
    """
    FileTest

    Object to manage tests carried out on files discovered
    during a library scan or inode event

    """

    def __init__(self, library_id: int):
        self.settings = config.Config()
        unmanic_logging = unlogger.UnmanicLogger.__call__()
        self.logger = unmanic_logging.get_logger(__class__.__name__)

        # Init plugins
        self.library_id = library_id
        self.plugin_handler = PluginsHandler()
        self.plugin_modules = self.plugin_handler.get_enabled_plugin_modules_by_type('library_management.file_test',
                                                                                     library_id=library_id)

        # List of filed tasks
        self.failed_paths = []

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def set_file(self):
        pass

    def file_failed_in_history(self, path):
        """
        Check if file has already failed in history

        :return:
        """
        # Fetch historical tasks
        history_logging = history.History()
        if not self.failed_paths:
            failed_tasks = history_logging.get_historic_tasks_list_with_source_probe(task_success=False)
            for task in failed_tasks:
                self.failed_paths.append(task.get('abspath'))
        if path in self.failed_paths:
            # That pathname was found in the results of failed historic tasks
            return True
        # No results were found matching that pathname
        return False

    def file_in_unmanic_ignore_lockfile(self, path):
        """
        Check if folder contains a '.unmanicignore' lockfile

        :return:
        """
        # Get file parent directory
        dirname = os.path.dirname(path)
        # Check if lockfile (.unmanicignore) exists
        unmanic_ignore_file = os.path.join(dirname, '.unmanicignore')
        if os.path.exists(unmanic_ignore_file):
            # Get file basename
            basename = os.path.basename(path)
            # Read the file and check for any entry with this file name
            with open(unmanic_ignore_file) as f:
                for line in f:
                    if basename in line:
                        return True
        return False

    def should_file_be_added_to_task_list(self, path):
        """
        Test if this file needs to be added to the task list

        :return:
        """
        return_value = None
        file_issues = []

        # TODO: Remove this
        if self.file_in_unmanic_ignore_lockfile(path):
            file_issues.append({
                'id':      'unmanicignore',
                'message': "File found in unmanic ignore file - '{}'".format(path),
            })
            return_value = False

        # Check if file has failed in history.
        if self.file_failed_in_history(path):
            file_issues.append({
                'id':      'blacklisted',
                'message': "File found already failed in history - '{}'".format(path),
            })
            return_value = False

        # Only run checks with plugins if other tests were not conclusive
        priority_score_modification = 0
        if return_value is None:
            # Set the initial data with just the priority score.
            data = {
                'priority_score': 0,
                'shared_info':    {},
            }
            # Run tests against plugins
            for plugin_module in self.plugin_modules:
                data['library_id'] = self.library_id
                data['path'] = path
                data['issues'] = deepcopy(file_issues)
                data['add_file_to_pending_tasks'] = None

                # Run plugin to update data
                if not self.plugin_handler.exec_plugin_runner(data, plugin_module.get('plugin_id'),
                                                              'library_management.file_test'):
                    continue

                # Append any file issues found during previous tests
                file_issues = data.get('issues')

                # Set the return_value based on the plugin results
                # If the add_file_to_pending_tasks returned an answer (True/False) then break the loop.
                # No need to continue.
                if data.get('add_file_to_pending_tasks') is not None:
                    return_value = data.get('add_file_to_pending_tasks')
                    break
            # Set the priority score modification
            priority_score_modification = data.get('priority_score', 0)

        return return_value, file_issues, priority_score_modification


class FileTesterThread(threading.Thread):
    def __init__(self, name, files_to_test, files_to_process, status_updates, library_id, event):
        super(FileTesterThread, self).__init__(name=name)
        self.settings = config.Config()
        self.logger = None
        self.event = event
        self.files_to_test = files_to_test
        self.files_to_process = files_to_process
        self.library_id = library_id
        self.status_updates = status_updates
        self.abort_flag = threading.Event()
        self.abort_flag.clear()

    def _log(self, message, message2='', level="info"):
        if not self.logger:
            unmanic_logging = unlogger.UnmanicLogger.__call__()
            self.logger = unmanic_logging.get_logger(self.name)
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        self.abort_flag.set()

    def run(self):
        self._log("Starting {}".format(self.name))
        file_test = FileTest(self.library_id)
        while not self.abort_flag.is_set():
            try:
                # Pending task queue has an item available. Fetch it.
                next_file = self.files_to_test.get_nowait()
                self.status_updates.put(next_file)
            except queue.Empty:
                self.event.wait(2)
                continue
            except Exception as e:
                self._log("Exception in fetching library scan result for path {}:".format(self.name), message2=str(e),
                            level="exception")

            # Test file to be added to task list. Add it if required
            try:
                result, issues, priority_score = file_test.should_file_be_added_to_task_list(next_file)
                # Log any error messages
                for issue in issues:
                    if type(issue) is dict:
                        self._log(issue.get('message'))
                    else:
                        self._log(issue)
                # If file needs to be added, then add it
                if result:
                    self.add_path_to_queue({
                        'path':           next_file,
                        'priority_score': priority_score,
                    })
            except UnicodeEncodeError:
                self._log("File contains Unicode characters that cannot be processed. Ignoring.", level="warning")
            except Exception as e:
                self._log("Exception testing file path in {}. Ignoring.".format(self.name), message2=str(e),
                            level="exception")

        self._log("Exiting {}".format(self.name))

    def add_path_to_queue(self, item):
        self.files_to_process.put(item)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.versioninfo.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     04 May 2020, (10:52 AM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""

import io
import os
import subprocess
import sys

import unmanic.metadata as version_info

DESCRIPTION = "TEST"


def name():
    return str(version_info.__name)


def version():
    if is_git_vcs():
        git_version_info = get_git_version_info()
        return git_version_info['short']
    else:
        return str(version_info.__version__)


def full_version():
    if is_git_vcs():
        git_version_info = get_git_version_info()
        return git_version_info['long']
    else:
        return str(version_info.__version__)


def description():
    return str(version_info.__description__)


def author():
    return str(version_info.__author)


def email():
    return str(version_info.__email)


def url():
    """
    Fetch the URL from the project version_info

    :return:
    """
    return str(version_info.__website__)


def branch_version():
    return version()[:3]


def is_pre_release():
    """
    Returns either true if this is a alpha or beta release

    :return:
    """
    full_version_string = full_version()
    return "alpha" in full_version_string.lower() or "beta" in full_version_string.lower()


def dev_status():
    """
    Returns the python module Development Status classifier
    based on if the version string is alpha/beta or stable

    :return:
    """
    full_version_string = full_version()
    if 'alpha' in full_version_string.lower():
        return 'Development Status :: 3 - Alpha'
    elif 'beta' in full_version_string or 'rc' in full_version_string.lower():
        return 'Development Status :: 4 - Beta'
    else:
        return 'Development Status :: 5 - Production/Stable'


def changes():
    """
    Extract part of changelog pertaining to version.

    :return:
    """
    _version = version_info.__version__
    with io.open(os.path.join(get_base_dir(), "CHANGES.txt"), 'r', encoding='utf8') as f:
        lines = []
        for line in f:
            if line.startswith('====='):
                if len(lines) > 1:
                    break
            if lines:
                lines.append(line)
            elif line.startswith(_version):
                lines.append(line)
    return ''.join(lines[:-1])


def get_base_dir():
    return os.path.abspath(os.path.dirname(sys.argv[0]))


def is_git_vcs():
    """
    Checks if the build is executed from a git VCS root

    :return:
    """
    if subprocess.call(["git", "branch"], stderr=subprocess.STDOUT, stdout=open(os.devnull, 'w')) != 0:
        return False
    else:
        return True


def get_git_version_info():
    """
    Parse version information from git.

    Returns a long and short version string:
        eg. short   - 0.0.1                         (full release)
        eg. short   - 0.0.1b2                       (beta release)
        eg. short   - 0.0.1b2.post3                 (beta release with commits since the last tag)
        eg. long    - 0.0.1~68b3db6                 (full release)
        eg. long    - 0.0.1-beta2~68b3db6           (beta release)
        eg. long    - 0.0.1-beta2+68b3db6           (beta release with commits since the last tag)
        eg. long    - 0.0.1-beta2~68b3db6+dirty     (beta release with uncommitted changes)
        eg. long    - 0.0.1-beta2+68b3db6+dirty     (beta release with commits since the last tag & uncommitted changes)

    :return:
    """
    # Fetch the last tag
    last_tag = subprocess.check_output(["git", "describe", "--tags", "--abbrev=0"]).strip().decode("utf-8")
    # Fetch the current commit ID
    current_commit = subprocess.check_output(["git", "rev-parse", "--verify", "--short", "HEAD"]).strip().decode("utf-8")

    # Create a long and short version string to return
    short_version_string = last_tag
    long_version_string = last_tag

    # Fetch the amount of commits since the last tag
    distance_since_last_tag = subprocess.check_output(
        ["git", "rev-list", last_tag + "..HEAD", "--count"]
    ).strip().decode("utf-8")

    # Normalize short version string (saves getting spammed with useless warnings from setuptools about it)
    if '-alpha' in short_version_string.lower():
        short_version_string = short_version_string.replace("-alpha", "a")
    elif '-beta' in short_version_string.lower():
        short_version_string = short_version_string.replace("-beta", "b")
    elif '-rc' in short_version_string.lower():
        short_version_string = short_version_string.replace("-rc", "rc")

    # Append a post tag if this is not a clean tagged build
    if int(distance_since_last_tag) > 0:
        # There are commits since the last tag
        # Modify the version strings
        short_version_string = '{}.post{}'.format(short_version_string, distance_since_last_tag)
        long_version_string = '{}+{}'.format(long_version_string, current_commit)
    else:
        long_version_string = '{}~{}'.format(long_version_string, current_commit)

    # Check if there are uncommitted changes on the directory
    git_diff_status = subprocess.check_output(
        "git diff-index --quiet HEAD -- || echo 'is_dirty'", shell=True
    ).strip().decode("utf-8")
    if git_diff_status == 'is_dirty':
        # There are commits since the last tag
        long_version_string = '{}+dirty'.format(long_version_string)

    return_dic = {
        'short': short_version_string,
        'long':  long_version_string
    }

    return return_dic



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.cli.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     21 Sep 2019, (2:31 PM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import subprocess

from ..exceptions.ffmpeg import FFMpegError
from ..exceptions.ffprobe import FFProbeError


def ffmpeg_cmd(params):
    """
    Execute a ffmpeg command subprocess and read the output

    :param params:
    :return:
    """
    command = ["ffmpeg"] + params

    pipe = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out, err = pipe.communicate()

    # Check for results
    try:
        raw_output = out.decode("utf-8")
    except Exception as e:
        raise FFMpegError(command, str(e))
    if pipe.returncode == 1 or 'error' in raw_output:
        raise FFMpegError(command, raw_output)
    if not raw_output:
        raise FFMpegError(command, 'No command output returned')

    return raw_output


def ffprobe_cmd(params):
    """
    Execute a ffprobe command subprocess and read the output

    :param params:
    :return:
    """
    command = ["ffprobe"] + params

    pipe = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    out, err = pipe.communicate()

    # Check for results
    try:
        raw_output = out.decode("utf-8")
    except Exception as e:
        raise FFProbeError(command, str(e))
    if pipe.returncode == 1 or 'error' in raw_output:
        raise FFProbeError(command, raw_output)
    if not raw_output:
        raise FFProbeError(command, 'No info found')

    return raw_output


def ffprobe_file(vid_file_path):
    """
    Returns a dictionary result from ffprobe command line prove of a file

    :param vid_file_path: The absolute (full) path of the video file, string.
    :return:
    """
    if type(vid_file_path) != str:
        raise Exception('Give ffprobe a full file path of the video')

    params = [
        "-loglevel", "quiet",
        "-print_format", "json",
        "-show_format",
        "-show_streams",
        "-show_error",
        vid_file_path
    ]

    # Check result
    results = ffprobe_cmd(params)
    try:
        info = json.loads(results)
    except Exception as e:
        raise FFProbeError(vid_file_path, str(e))

    return info


def ffmpeg_version_info():
    """
    Returns a dictionary result of the current FFMPEG versions

    :return:
    """
    params = [
        "-loglevel", "quiet",
        "-print_format", "json",
        "-show_versions",
    ]

    results = ffprobe_cmd(params)
    try:
        info = json.loads(results)
    except Exception as e:
        raise FFProbeError("ffmpeg_version_info function", str(e))

    return info


def ffmpeg_available_encoders():
    """
    Return the raw output of encoders supported by ffmpeg
      Encoders:
        V..... = Video
        A..... = Audio
        S..... = Subtitle
        .F.... = Frame-level multithreading
        ..S... = Slice-level multithreading
        ...X.. = Codec is experimental
        ....B. = Supports draw_horiz_band
        .....D = Supports direct rendering method 1

    :return:
    """
    params = [
        "-loglevel",
        "quiet",
        "-encoders",
    ]

    return ffmpeg_cmd(params)


def ffmpeg_available_decoders():
    """
    Return the raw output of decoders supported by ffmpeg
      Decoders:
        V..... = Video
        A..... = Audio
        S..... = Subtitle
        .F.... = Frame-level multithreading
        ..S... = Slice-level multithreading
        ...X.. = Codec is experimental
        ....B. = Supports draw_horiz_band
        .....D = Supports direct rendering method 1


    :return:
    """
    params = [
        "-loglevel",
        "quiet",
        "-decoders",
    ]

    return ffmpeg_cmd(params)


def ffmpeg_available_hw_acceleration_methods():
    """
    Return the raw output of hardware accelration methods supported by ffmpeg
      Hardware acceleration methods:
      vdpau
      cuda
      vaapi


    :return:
    """
    params = [
        "-loglevel",
        "quiet",
        "-hwaccels",
    ]

    return ffmpeg_cmd(params)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.common.py
 
    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     06 Dec 2018, (7:21 AM)
 
    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved
 
           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:
  
           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.
  
           THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import copy
import datetime
import hashlib
import os
import random
import string
import shutil


def get_home_dir():
    # Attempt to get the HOME_DIR environment variable
    home_dir = os.environ.get('HOME_DIR')
    # If HOME_DIR is unset, empty, or specifically set to use the home directory
    if not home_dir:
        # Expand the tilde to the user's home directory
        home_dir = os.path.expanduser("~")
    else:
        # For any value of HOME_DIR, ensure tilde and relative paths are correctly handled
        # os.path.expanduser will handle tilde but won't affect absolute paths
        # os.path.abspath will convert a relative path to an absolute path
        home_dir = os.path.abspath(os.path.expanduser(home_dir))
    return home_dir


def get_default_root_path():
    root = os.path.join(os.sep)
    if os.name == "nt":
        root = os.path.join('c:', os.sep)
    return root


def get_default_library_path():
    library_path = os.path.join(get_default_root_path(), 'library')
    # Windows set the default library directory into %USERPROFILE%\Documents
    if os.name == "nt":
        library_path = os.path.join(os.path.expandvars(r'%USERPROFILE%'), 'Documents')
    return library_path


def get_default_cache_path():
    cache_path = os.path.join(get_default_root_path(), 'tmp', 'unmanic')
    # Windows set the default temp directory into %LOCALAPPDATA%\Temp\Unmanic
    if os.name == "nt":
        cache_path = os.path.join(os.path.expandvars(r'%LOCALAPPDATA%\Temp'), 'Unmanic')
    return cache_path


def format_message(message, message2=''):
    message = str(message)
    if message2:
        # Message2 can support other objects:
        if isinstance(message2, str):
            message = "%s - %s" % (message, str(message2))
        elif isinstance(message2, dict) or isinstance(message2, list):
            import pprint
            message2 = pprint.pformat(message2, indent=1)
            message = "%s \n%s" % (message, str(message2))
        else:
            message = "%s - %s" % (message, str(message2))
    message = "[FORMATTED] - %s" % message
    return message


def make_timestamp_human_readable(ts):
    """
    Accept a unix timestamp, return a human readable timedelta string.

    :param ts: a datetime, timedelta, or timestamp (integer / float) object
    :returns: Human readable timedelta string (Str)
    """
    units = ("year", "day", "hour", "minute", "second", "millisecond", "microsecond")
    precision = 1
    past_tense = "{} ago"
    future_tense = "in {}"

    # Get datetime from ts string
    dt = datetime.datetime.fromtimestamp(ts)
    delta = datetime.datetime.now(tz=dt.tzinfo) - dt

    # Determine if this is past or future tense
    the_tense = future_tense if delta < datetime.timedelta(0) else past_tense

    # Create a dictionary of units
    delta = abs(delta)
    d = {
        "year":        int(delta.days / 365),
        "day":         int(delta.days % 365),
        "hour":        int(delta.seconds / 3600),
        "minute":      int(delta.seconds / 60) % 60,
        "second":      delta.seconds % 60,
        "millisecond": delta.microseconds / 1000,
        "microsecond": delta.microseconds % 1000,
    }

    human_readable_list = []
    count = 0

    # Start building up the output in the human readable list.
    for unit in units:
        if count >= precision:
            break  # met precision
        if d[unit] == 0:
            continue  # skip 0's
        else:
            s = "" if d[unit] == 1 else "s"  # handle plurals
            human_readable_list.append("{} {}{}".format(d[unit], unit, s))
        count += 1

    return the_tense.format(", ".join(human_readable_list))


def ensure_dir(file_path):
    directory = os.path.dirname(file_path)
    if not os.path.exists(directory):
        os.makedirs(directory)


def time_string_to_seconds(time_string):
    pt = datetime.datetime.strptime(time_string, '%H:%M:%S.%f')
    return pt.second + pt.minute * 60 + pt.hour * 3600


def tail(f, n, offset=0):
    """Reads a n lines from f with an offset of offset lines."""
    avg_line_length = 153
    to_read = n + offset
    while 1:
        try:
            f.seek(-(avg_line_length * to_read), 2)
            while f.read(1) != b'\n':
                f.seek(-2, os.SEEK_CUR)
        except IOError:
            f.seek(0)
        pos = f.tell()
        lines = f.read().splitlines()
        if len(lines) >= to_read or pos == 0:
            return lines
        avg_line_length *= 1.3


def touch(fname, mode=0o666, dir_fd=None, **kwargs):
    """Touch a file. If it does not exist, create it."""
    flags = os.O_CREAT | os.O_APPEND
    with os.fdopen(os.open(fname, flags=flags, mode=mode, dir_fd=dir_fd)) as f:
        os.utime(f.fileno() if os.utime in os.supports_fd else fname,
                 dir_fd=None if os.supports_fd else dir_fd, **kwargs)


def clean_files_in_cache_dir(cache_directory):
    """This will completely wipe all contents from a directory"""
    if os.path.exists(cache_directory):
        for root, subFolders, files in os.walk(cache_directory):
            root_bn = os.path.basename(root)
            if root_bn.startswith("unmanic_file_conversion-"):
                try:
                    print("Clearing cache path - {}".format(root))
                    shutil.rmtree(root)
                except Exception as e:
                    print("Exception while clearing cache path - {}".format(str(e)))
            elif root_bn.startswith("unmanic_remote_pending_library-"):
                try:
                    print("Clearing remote library cache path - {}".format(root))
                    shutil.rmtree(root)
                except Exception as e:
                    print("Exception while clearing remote library cache path - {}".format(str(e)))


def random_string(string_length=5):
    """Generate a random string of fixed length"""
    letters = string.ascii_lowercase
    return ''.join(random.choice(letters) for i in range(string_length))


def json_dump_to_file(json_data, out_file, check=True, rollback_on_fail=True):
    """Dump json data to a file. Optionally checks that the output json data is valid"""
    import json
    import time
    import tempfile
    import shutil

    result = {
        'errors':  [],
        'success': False
    }

    # If check param is flagged and there already exists a out file, create a temporary backup

    if rollback_on_fail and os.path.exists(out_file):
        temp_dir = tempfile.gettempdir()
        temp_path = os.path.join(temp_dir, 'json_dump_to_file_backup-{}'.format(time.time()))
        try:
            shutil.copy2(out_file, temp_path)
            result['temp_path'] = temp_path
        except Exception as e:
            result['success'] = False
            result['errors'].append("Failed to create temporary file - {}".format(str(e)))

    # Write data to out_file
    try:
        with open(out_file, 'w') as outfile:
            json.dump(json_data, outfile, sort_keys=True, indent=4)
        result['success'] = True
    except Exception as e:
        result['success'] = False
        result['errors'].append("Exception in writing to file: {}".format(str(e)))

    # If check param is flagged, ensure json data exists in the output file
    if check:
        try:
            with open(out_file) as infile:
                data = json.load(infile)
        except Exception as e:
            result['success'] = False
            result['errors'].append("JSON file invalid - {}".format(e))

    # If data save was unsuccessful and the rollback_on_fail param is flagged
    #   and there is a temp file set, roll back to old file
    if not result.get('success') and result.get('temp_path') and rollback_on_fail:
        try:
            os.remove(out_file)
            shutil.copy2(result.get('temp_path'), out_file)
            os.remove(result.get('temp_path'))
        except Exception as e:
            result['success'] = False
            result['errors'].append("Exception while restoring original file file: {}".format(str(e)))

    return result


def extract_video_codecs_from_file_properties(file_properties: dict):
    """
    Read a dictionary of file properties
    Extract a list of video codecs from the video streams

    :param file_properties:
    :return:
    """
    codecs = []
    for stream in file_properties['streams']:
        if stream['codec_type'] == 'video':
            codecs.append(stream['codec_name'])
    return codecs


def get_file_checksum(path):
    """
    Read a checksum of a file.

    Rather than opening the whole file in memory, open it in chunks.
    This is slightly slower, but allows working on systems with limited memory.

    :param path:
    :return:
    """
    file_hash = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b''):
            file_hash.update(chunk)
    return copy.copy(file_hash.hexdigest())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.pluginscli.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     15 Mar 2021, (12:05 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import json
import logging
import os
import re
import shutil
import subprocess

import inquirer
import requests

from . import plugin_types

from unmanic import config
from unmanic.libs import unlogger, common
from unmanic.libs.plugins import PluginsHandler
from unmanic.libs.unplugins import PluginExecutor

home_directory = common.get_home_dir()
dev_cache_directory = os.path.join(home_directory, '.unmanic', 'dev', 'cache')
dev_library_directory = os.path.join(home_directory, '.unmanic', 'dev', 'library')

menus = {
    "main":          [
        inquirer.List(
            'cli_action',
            message="What would you like to do?",
            choices=[
                'List all installed plugins',
                'Test plugins',
                'Create new plugin',
                'Reload all plugins from disk',
                'Remove plugin',
                'Install test data',
                'Exit',
            ],
        ),
    ],
    "create_plugin": [
        inquirer.Text('plugin_id', message="What's the plugin's id"),
        inquirer.Text('plugin_name', message="What's the plugin's name"),
    ],
}


class BColours:
    HEADER = '\033[44m'
    SUBHEADER = '\033[94m'
    SECTION = '\033[96m'
    RESULTS = '\033[39m'
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    WARNING = '\033[93m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_table(table_data, col_list=None, sep='\uFFFA', max_col_width=9):
    """
    Pretty print a list of dictionaries (myDict) as a dynamically sized table.
    If column names (col_list) aren't specified, they will show in random order.

    Author: Thierry Husson

    """
    if not col_list:
        col_list = list(table_data[0].keys() if table_data else [])
    my_list = [col_list]  # 1st row = header
    for item in table_data:
        my_list.append([str(item[col] or '') for col in col_list])
    original_col_size = [max(map(len, (sep.join(col)).split(sep))) for col in zip(*my_list)]
    col_size = []
    for col in original_col_size:
        if col > max_col_width:
            col = max_col_width
        col_size.append(col)
    format_str = ' | '.join(["{{:<{}}}".format(i) for i in col_size])
    line = format_str.replace(' | ', '-+-').format(*['-' * i for i in col_size])
    item = my_list.pop(0)
    line_done = False
    while my_list:
        if all(not i for i in item):
            item = my_list.pop(0)
            if line and (sep != '\uFFFA' or not line_done):
                print(line)
                line_done = True
        row = [i[:max_col_width].split(sep, 1) for i in item]
        print(format_str.format(*[i[0] for i in row]))
        item = [i[1] if len(i) > 1 else '' for i in row]


def install_npm_modules(plugin_path):
    package_file = os.path.join(plugin_path, 'package.json')
    if not os.path.exists(package_file):
        return
    subprocess.call(['npm', 'install'], cwd=plugin_path)
    subprocess.call(['npm', 'run', 'build'], cwd=plugin_path)


def install_plugin_requirements(plugin_path):
    requirements_file = os.path.join(plugin_path, 'requirements.txt')
    install_target = os.path.join(plugin_path, 'site-packages')
    if not os.path.exists(requirements_file):
        return
    import pip
    pip.main(['install', '--upgrade', '-r', requirements_file, '--target={}'.format(install_target)])


class PluginsCLI(object):

    def __init__(self, plugins_directory=None):
        # Read settings
        self.settings = config.Config()

        # Set plugins directory
        if not plugins_directory:
            home_directory = common.get_home_dir()
            plugins_directory = os.path.join(home_directory, '.unmanic', 'plugins')
        self.plugins_directory = plugins_directory
        unmanic_logging = unlogger.UnmanicLogger.__call__()
        unmanic_logging.disable_file_handler(debugging=True)
        unmanic_logging.stream_handler.setFormatter(
            logging.Formatter(
                '        - {}%(asctime)s:%(levelname)s:%(name)s - %(message)s{}'.format(BColours.RESULTS, BColours.ENDC),
                datefmt='%Y-%m-%dT%H:%M:%S'
            )
        )
        unmanic_logging.enable_debugging()
        self.logger = unmanic_logging.get_logger(__class__.__name__)

        self.test_data_modifiers = {
            "{cache_path}":    dev_cache_directory,
            "{library_path}":  dev_library_directory,
            "{test_file_in}":  "Big_Buck_Bunny_1080_10s_30MB_h264.mkv",
            "{test_file_out}": "Big_Buck_Bunny_1080_10s_30MB_h264-1616571944.7296877-WORKING-1.mkv"
        }

    def _log(self, message, message2='', level="info"):
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def create_new_plugins(self):
        plugin_details = inquirer.prompt(menus.get('create_plugin'))

        # Ensure results are not empty
        if not plugin_details.get('plugin_name') or not plugin_details.get('plugin_id'):
            print("ERROR! Invalid input.")
            return

        # Ensure plugin ID has only underscore and a-z, 0-9
        plugin_details['plugin_id'] = re.sub('[^0-9a-zA-Z]+', '_', plugin_details.get('plugin_id'))
        # Ensure plugin ID is lower case
        plugin_details['plugin_id'] = plugin_details.get('plugin_id').lower()

        # Get list of plugin types
        all_plugin_types = plugin_types.get_all_plugin_types()

        # Build choice selection list from installed plugins
        plugin_details_by_runner = {}
        choices = []
        for plugin_type in all_plugin_types:
            choices.append(all_plugin_types[plugin_type].get('name'))
            plugin_details_by_runner[all_plugin_types[plugin_type].get('name')] = all_plugin_types[plugin_type]

        # Generate menu menu
        print()
        print('INFO: https://docs.unmanic.app/docs/plugins/writing_plugins/plugin_runner_types')
        plugin_runners_inquirer = inquirer.List(
            'selected_plugin',
            message="Which Plugin runner will be used?",
            choices=choices,
        )

        # Prompt for selection of Plugin by ID
        runner_selection = inquirer.prompt([plugin_runners_inquirer])

        # Fetch plugin type details from selection
        plugin_type_details = plugin_details_by_runner[runner_selection.get('selected_plugin')]
        selected_plugin_runner = plugin_type_details.get('runner')
        selected_plugin_runner_docstring = plugin_type_details.get('runner_docstring')

        # Create new plugin path
        new_plugin_path = os.path.join(self.plugins_directory, plugin_details.get('plugin_id'))
        if not os.path.exists(new_plugin_path):
            os.makedirs(new_plugin_path)

        # Create main python file template
        main_plugin_template = [
            "#!/usr/bin/env python3",
            "# -*- coding: utf-8 -*-",
            "",
            "import logging",
            "",
            "from unmanic.libs.unplugins.settings import PluginSettings",
            "",
            "# Configure plugin logger",
            "logger = logging.getLogger(\"Unmanic.Plugin.{}\")".format(plugin_details.get('plugin_id')),
            "",
            "",
            "class Settings(PluginSettings):",
            "    settings = {}",
            "",
            "    def __init__(self, *args, **kwargs):",
            "        super(Settings, self).__init__(*args, **kwargs)",
            "",
            "",
        ]

        # Create runner function template
        runner_template = [
            'def {}(data):'.format(selected_plugin_runner),
            '    """{}'.format(selected_plugin_runner_docstring),
            '    """',
            '    return',
        ]

        # Write above templates to main python file
        main_python_file = os.path.join(new_plugin_path, 'plugin.py')
        if not os.path.exists(main_python_file):
            with open(main_python_file, 'a') as outfile:
                # Write out main template
                for template_line in main_plugin_template:
                    outfile.write("{}\n".format(template_line))
                # Write out runner function template
                for template_line in runner_template:
                    outfile.write("{}\n".format(template_line))

        # Write plugin info.json
        info_file = os.path.join(new_plugin_path, 'info.json')
        plugin_info = {
            "id":            plugin_details.get('plugin_id'),
            "name":          plugin_details.get('plugin_name'),
            "author":        "",
            "version":       "0.0.1",
            "tags":          "",
            "description":   "",
            "icon":          "",
            "priorities":    {
                selected_plugin_runner: 0
            },
            "compatibility": [PluginsHandler.version]
        }
        if not os.path.exists(info_file):
            with open(info_file, 'w') as outfile:
                json.dump(plugin_info, outfile, sort_keys=True, indent=4)

        # Create requirements.txt file
        common.touch(os.path.join(new_plugin_path, 'requirements.txt'))

        # Create Plugin .gitignore
        plugin_gitignore = os.path.join(new_plugin_path, '.gitignore')
        gitignore_template = [
            '**/__pycache__',
            '*.py[cod]',
            '**/site-packages',
            'settings.json',
        ]
        if not os.path.exists(plugin_gitignore):
            with open(plugin_gitignore, 'a') as outfile:
                for template_line in gitignore_template:
                    outfile.write("{}\n".format(template_line))

        # Insert plugin details to DB
        try:
            plugin_info['plugin_id'] = plugin_info.get('id')
            PluginsHandler.write_plugin_data_to_db(plugin_info, new_plugin_path)
        except Exception as e:
            print("Exception while saving plugin info to DB. - {}".format(str(e)))
            return

        print("Plugin created - '{}'".format((plugin_details.get('plugin_id'))))

    def reload_plugin_from_disk(self):
        # Fetch list of installed plugins
        plugin_results = self.__get_installed_plugins()

        # Build choice selection list from installed plugins
        for plugin in plugin_results:
            print("Reloading Plugin - '{}'".format(plugin.get('plugin_id')))
            plugin_path = os.path.join(self.plugins_directory, plugin.get('plugin_id'))
            # Read plugin info.json
            info_file = os.path.join(plugin_path, 'info.json')
            with open(info_file) as json_file:
                plugin_info = json.load(json_file)

            # Insert plugin details to DB
            try:
                plugin_info['plugin_id'] = plugin_info.get('id')
                PluginsHandler.write_plugin_data_to_db(plugin_info, plugin_path)
            except Exception as e:
                print("Exception while saving plugin info to DB. - {}".format(str(e)))
                return

            install_plugin_requirements(plugin_path)
            install_npm_modules(plugin_path)
            print()
        print()

    def remove_plugin(self):
        # Fetch list of installed plugins
        plugin_results = self.__get_installed_plugins()

        # Build choice selection list from installed plugins
        table_ids = {}
        choices = []
        for plugin in plugin_results:
            choices.append(plugin.get('plugin_id'))
            table_ids[plugin.get('plugin_id')] = plugin.get('id')
        # Append a "return" option
        choices.append('Go Back')

        # Generate menu menu
        remove_plugin_inquirer = inquirer.List(
            'cli_action',
            message="Which Plugin would you like to remove?",
            choices=choices,
        )

        # Prompt for selection of Plugin by ID
        selection = inquirer.prompt([remove_plugin_inquirer])

        # If the 'Go Back' option was given, just return to previous menu
        if not selection or selection.get('cli_action') == "Go Back":
            return

        # Remove the selected Plugin by ID
        plugin_table_id = table_ids[selection.get('cli_action')]
        plugins = PluginsHandler()
        plugins.uninstall_plugins_by_db_table_id([plugin_table_id])
        print()

    def list_installed_plugins(self):
        plugin_results = self.__get_installed_plugins()
        print_table(plugin_results)
        print()
        print()

    @staticmethod
    def __get_installed_plugins(plugin_id=None):
        plugins = PluginsHandler()
        order = [
            {
                "column": 'plugin_id',
                "dir":    'asc',
            },
        ]
        if plugin_id:
            return plugins.get_plugin_list_filtered_and_sorted(order=order, start=0, length=None, plugin_id=plugin_id)
        return plugins.get_plugin_list_filtered_and_sorted(order=order, start=0, length=None)

    def test_installed_plugins(self, plugin_id=None):
        """
        Test all plugin runners for correct return data.
        If plugin_id is provided, only the matching plugin will be tested.

        :param plugin_id:
        :return:
        """
        plugin_executor = PluginExecutor()

        plugin_results = self.__get_installed_plugins(plugin_id=plugin_id)
        for plugin_result in plugin_results:
            # plugin_runners = plugin_executor.get_plugin_runners('worker.process_item')
            print("{1}Testing plugin: '{0}'{2}".format(plugin_result.get("name"), BColours.HEADER, BColours.ENDC))
            plugin_id = plugin_result.get("plugin_id")

            # Reload the plugin
            plugin_executor.reload_plugin_module(plugin_id)

            # Test Plugin runners
            print("  {0}Testing runners{1}".format(BColours.SUBHEADER, BColours.ENDC))
            plugin_types_in_plugin = plugin_executor.get_all_plugin_types_in_plugin(plugin_id)
            if not plugin_types_in_plugin:
                error = "No runners found in plugin"
                print("  -- {1}FAILED: {0}{2}".format(error, BColours.FAIL, BColours.ENDC))
                print()
            else:
                for plugin_type_in_plugin in plugin_types_in_plugin:
                    print("    {1}{0}{2}".format(plugin_type_in_plugin, BColours.SECTION, BColours.ENDC))
                    errors = plugin_executor.test_plugin_runner(plugin_id, plugin_type_in_plugin,
                                                                test_data_modifiers=self.test_data_modifiers)
                    if errors:
                        for error in errors:
                            print("        -- {1}FAILED: {0}{2}".format(error, BColours.FAIL, BColours.ENDC))
                    else:
                        print("        -- {}PASSED{} --".format(BColours.OKGREEN, BColours.ENDC))
                    print()

            # Test Plugin settings
            print("  {0}Testing settings{1}".format(BColours.SUBHEADER, BColours.ENDC))
            errors, plugin_settings = plugin_executor.test_plugin_settings(plugin_id)
            print("    {}Plugin settings schema{}".format(BColours.SECTION, BColours.ENDC))
            if errors:
                for error in errors:
                    print("        -- {1}FAILED: {0}{2}".format(error, BColours.FAIL, BColours.ENDC))
            else:
                formatted_plugin_settings = json.dumps(plugin_settings, indent=1)
                formatted_plugin_settings = formatted_plugin_settings.replace('\n', '\n' + '                    ')
                print("        - {1}Settings: {0}{2}".format(formatted_plugin_settings, BColours.RESULTS, BColours.ENDC))
                print("        -- {}PASSED{} --".format(BColours.OKGREEN, BColours.ENDC))
            print()
            print()

    def test_plugins(self):
        plugin_results = self.__get_installed_plugins()

        # Generate menu choices
        all_plugin_details = {}
        choices = ["Configure Testdata", "Test All Plugins"]
        for plugin_details in plugin_results:
            choices.append(plugin_details.get('plugin_id'))
            all_plugin_details[plugin_details.get('plugin_id')] = plugin_details
        choices.append("Go Back")

        print()
        default_selection = None
        while True:
            plugin_test_inquirer = inquirer.List(
                'selected_plugin',
                message="Which plugin would you like to test?",
                choices=choices,
                default=default_selection,
            )
            selection = inquirer.prompt([plugin_test_inquirer])

            # If the 'Go Back' option was given, just return to previous menu
            if not selection or selection.get('selected_plugin') == "Go Back":
                return

            # Configure test file
            if selection.get('selected_plugin') == "Configure Testdata":
                self.configure_test_data()
                continue

            # If 'Test All Plugins' was selected, then run tests against all plugins
            if selection.get('selected_plugin') == "Test All Plugins":
                self.test_installed_plugins()
                continue

            # Get the details for the selected plugin
            selected_plugin_details = all_plugin_details[selection.get('selected_plugin')]
            # Set that selection as the default for the next time
            default_selection = selection.get('selected_plugin')
            # Test that plugin
            self.test_installed_plugins(plugin_id=selected_plugin_details.get('plugin_id'))

    def configure_test_data(self):

        test_files = []
        for (dir_path, dir_names, file_names) in os.walk(dev_library_directory):
            test_files.extend(file_names)
        print(test_files)

        # Update test file
        print()
        print()
        test_files_inquirer = inquirer.List(
            'selected_file',
            message="Which Plugin runner will be used?",
            choices=test_files,
        )
        runner_selection = {}
        runner_selection = {
            **inquirer.prompt([test_files_inquirer]),
            **runner_selection
        }
        self.test_data_modifiers['{test_file_in}'] = runner_selection.get('selected_file')
        split_file_in = os.path.splitext(runner_selection.get('selected_file'))
        self.test_data_modifiers['{test_file_out}'] = "{}-{}{}".format(split_file_in[0], "WORKING-1", split_file_in[1])

    def install_test_data(self):
        sample_files = {
            "Big_Buck_Bunny_1080_10s_30MB_h264.mkv": "https://test-videos.co.uk/vids/bigbuckbunny/mkv/1080/Big_Buck_Bunny_1080_10s_30MB.mkv",
            "Big_Buck_Bunny_1080_10s_30MB_h264.mp4": "https://test-videos.co.uk/vids/bigbuckbunny/mp4/h264/1080/Big_Buck_Bunny_1080_10s_30MB.mp4",
            "Big_Buck_Bunny_1080_10s_30MB_av1.mp4":  "https://test-videos.co.uk/vids/bigbuckbunny/mp4/av1/1080/Big_Buck_Bunny_1080_10s_30MB.mp4",
            "sample-12s.mp3":                        "https://download.samplelib.com/mp3/sample-12s.mp3",
        }
        if not os.path.exists(dev_cache_directory):
            os.makedirs(dev_cache_directory)
        if not os.path.exists(dev_library_directory):
            os.makedirs(dev_library_directory)
        for key in sample_files:
            library_file = os.path.join(dev_library_directory, key)
            file_url = sample_files.get(key)
            print()
            print("Downloading sample file: '{}'".format(file_url))
            with requests.get(file_url, stream=True) as r:
                r.raise_for_status()
                with open(library_file, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=None):
                        if chunk:
                            f.write(chunk)

            split_file_in = os.path.splitext(key)
            cache_file = os.path.join(dev_cache_directory, "{}-{}{}".format(split_file_in[0], "WORKING-1", split_file_in[1]))
            shutil.copyfile(library_file, cache_file)

    def main(self, arg):
        switcher = {
            'List all installed plugins':   'list_installed_plugins',
            'Test plugins':                 'test_plugins',
            'Create new plugin':            'create_new_plugins',
            'Reload all plugins from disk': 'reload_plugin_from_disk',
            'Remove plugin':                'remove_plugin',
            'Install test data':            'install_test_data',
        }
        function = switcher.get(arg, None)
        if function:
            getattr(self, function)()
        else:
            self._log("Invalid selection")
            return

    def run(self):
        print()
        while True:
            selection = inquirer.prompt(menus.get('main'))
            if not selection or selection.get('cli_action') == "Exit":
                break
            self.main(selection.get('cli_action'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
    unmanic.libraryscanner.py

    Written by:               Josh.5 <jsunnex@gmail.com>
    Date:                     20 Aug 2021, (5:37 PM)

    Copyright:
           Copyright (C) Josh Sunnex - All Rights Reserved

           Permission is hereby granted, free of charge, to any person obtaining a copy
           of this software and associated documentation files (the "Software"), to deal
           in the Software without restriction, including without limitation the rights
           to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
           copies of the Software, and to permit persons to whom the Software is
           furnished to do so, subject to the following conditions:

           The above copyright notice and this permission notice shall be included in all
           copies or substantial portions of the Software.

           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
           EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
           MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
           IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
           DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
           OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
           OR OTHER DEALINGS IN THE SOFTWARE.

"""
import gc
import json
import os
import queue
import threading
import time

import psutil
import schedule

from unmanic import config
from unmanic.libs import common, unlogger
from unmanic.libs.filetest import FileTesterThread
from unmanic.libs.library import Library
from unmanic.libs.plugins import PluginsHandler


class LibraryScannerManager(threading.Thread):
    def __init__(self, data_queues, event):
        super(LibraryScannerManager, self).__init__(name='LibraryScannerManager')
        self.interval = 0
        self.firstrun = True
        self.data_queues = data_queues
        self.settings = config.Config()
        self.logger = None
        self.event = event
        self.scheduledtasks = data_queues["scheduledtasks"]
        self.library_scanner_triggers = data_queues["library_scanner_triggers"]
        self.abort_flag = threading.Event()
        self.abort_flag.clear()
        self.scheduler = schedule.Scheduler()

        self.file_test_managers = {}
        self.files_to_test = queue.Queue()
        self.files_to_process = queue.Queue()

    def _log(self, message, message2='', level="info"):
        if not self.logger:
            unmanic_logging = unlogger.UnmanicLogger.__call__()
            self.logger = unmanic_logging.get_logger(self.name)
        message = common.format_message(message, message2)
        getattr(self.logger, level)(message)

    def stop(self):
        self.abort_flag.set()
        # Stop all child threads
        self.stop_all_file_test_managers()

    def abort_is_set(self):
        # Check if the abort flag is set
        if self.abort_flag.is_set():
            # Return True straight away if it is
            return True
        # Sleep for a fraction of a second to prevent CPU pinning
        self.event.wait(.1)
        # Return False
        return False

    def run(self):
        self._log("Starting LibraryScanner Monitor loop")
        while not self.abort_is_set():
            self.event.wait(1)

            # Main loop to configure the scheduler
            if int(self.settings.get_schedule_full_scan_minutes()) != self.interval:
                self.interval = int(self.settings.get_schedule_full_scan_minutes())
            if self.interval and self.interval != 0:
                self._log("Setting LibraryScanner schedule to scan every {} mins...".format(self.interval))
                # Configure schedule
                self.scheduler.every(self.interval).minutes.do(self.scheduled_job)
                # Register application
                self.register_unmanic()

                # First run the task
                if self.settings.get_run_full_scan_on_start() and self.firstrun:
                    self._log("Running LibraryScanner on start")
                    self.scheduled_job()
                self.firstrun = False

                # Then loop and wait for the schedule
                while not self.abort_is_set():
                    # Delay for 1 second before checking again.
                    self.event.wait(1)

                    # Check if a manual library scan was triggered
                    try:
                        if not self.library_scanner_triggers.empty():
                            trigger = self.library_scanner_triggers.get_nowait()
                            if trigger == "library_scan":
                                self.scheduled_job()
                                break
                    except queue.Empty:
                        continue
                    except Exception as e:
                        self._log("Exception in retrieving library scanner trigger {}:".format(self.name), message2=str(e),
                                  level="exception")

                    # Check if library scanner is enabled
                    if not self.settings.get_enable_library_scanner():
                        # The library scanner is not enabled. Dont run anything
                        continue

                    # Check if scheduled task is due
                    self.scheduler.run_pending()

                    # If the settings have changed, then break this loop and clear
                    # the scheduled job resetting to the new interval
                    if int(self.settings.get_schedule_full_scan_minutes()) != self.interval:
                        self._log("Resetting LibraryScanner schedule")
                        break
                self.scheduler.clear()

        self._log("Leaving LibraryScanner Monitor loop...")

    def scheduled_job(self):
        """
        Function called by the scheduled task

        :return:
        """
        if not self.system_configuration_is_valid():
            self._log("Skipping library scanner due invalid system configuration.", level='warning')
            return

        # For each configured library, check if a library scan is required
        no_libraries_configured = True
        for lib_info in Library.get_all_libraries():
            no_libraries_configured = False
            try:
                library = Library(lib_info['id'])
            except Exception as e:
                self._log("Unable to fetch library config for ID {}".format(lib_info['id']), level='exception')
                continue
            # Check if the library is configured for remote files only
            if library.get_enable_remote_only():
                # This library is configured to receive remote files only... Never run a library scan on it
                continue
            # Check if library scanner is enabled on any library
            if library.get_enable_scanner():
                # Run library scan
                self._log("Running full library scan on library '{}'".format(library.get_name()))
                self.scan_library_path(library.get_path(), library.get_id())
        if no_libraries_configured:
            self._log("No libraries are configured to run a library scan")

    def system_configuration_is_valid(self):
        """
        Check and ensure the system configuration is correct for running

        :return:
        """
        valid = True
        plugin_handler = PluginsHandler()
        if plugin_handler.get_incompatible_enabled_plugins(self.data_queues.get('frontend_messages')):
            valid = False
        if not Library.within_library_count_limits(self.data_queues.get('frontend_messages')):
            valid = False
        return valid

    def add_path_to_queue(self, pathname, library_id, priority_score):
        self.scheduledtasks.put({
            'pathname':       pathname,
            'library_id':     library_id,
            'priority_score': priority_score,
        })

    def start_results_manager_thread(self, manager_id, status_updates, library_id):
        manager = FileTesterThread("FileTesterThread-{}".format(manager_id), self.files_to_test,
                                   self.files_to_process, status_updates, library_id, self.event)
        manager.daemon = True
        manager.start()
        self.file_test_managers[manager_id] = manager

    def stop_all_file_test_managers(self):
        for manager_id in self.file_test_managers:
            self.file_test_managers[manager_id].abort_flag.set()

    def scan_library_path(self, library_path, library_id):
        """
        Run a scan of the given library path

        :param library_path:
        :param library_id:
        :return:
        """
        if not os.path.exists(library_path):
            self._log("Path does not exist - '{}'".format(library_path), level="warning")
            return
        if self.settings.get_debugging():
            self._log("Scanning directory - '{}'".format(library_path), level="debug")

        # Push status notification to frontend
        frontend_messages = self.data_queues.get('frontend_messages')

        # Start X number of FileTesterThread threads
        concurrent_file_testers = self.settings.get_concurrent_file_testers()
        status_updates = queue.Queue()
        self.file_test_managers = {}
        for results_manager_id in range(int(concurrent_file_testers)):
            self.start_results_manager_thread(results_manager_id, status_updates, library_id)

        start_time = time.time()

        frontend_messages.update(
            {
                'id':      'libraryScanProgress',
                'type':    'status',
                'code':    'libraryScanProgress',
                'message': "Scanning directory - '{}'".format(library_path),
                'timeout': 0
            }
        )

        follow_symlinks = self.settings.get_follow_symlinks()
        total_file_count = 0
        current_file = ''
        percent_completed_string = ''
        for root, subFolders, files in os.walk(library_path, followlinks=follow_symlinks):
            if self.abort_flag.is_set():
                break
            if self.settings.get_debugging():
                self._log(json.dumps(files, indent=2), level="debug")
            # Add all files in this path that match our container filter
            for file_path in files:
                if self.abort_flag.is_set():
                    break

                # Place file's full path in queue to be tested
                self.files_to_test.put(os.path.join(root, file_path))
                total_file_count += 1

                # Update status messages while fetching file list
                if not status_updates.empty():
                    current_file = status_updates.get()
                    percent_completed_string = 'Testing: {}'.format(current_file)
                    frontend_messages.update(
                        {
                            'id':      'libraryScanProgress',
                            'type':    'status',
                            'code':    'libraryScanProgress',
                            'message': percent_completed_string,
                            'timeout': 0
                        }
                    )

        # Loop while waiting for all threads to finish
        double_check = 0
        while not self.abort_flag.is_set():
            frontend_messages.update(
                {
                    'id':      'libraryScanProgress',
                    'type':    'status',
                    'code':    'libraryScanProgress',
                    'message': percent_completed_string,
                    'timeout': 0
                }
            )
            # Check if all files have been tested
            if self.files_to_test.empty() and self.files_to_process.empty() and status_updates.empty():
                percent_completed_string = '100%'
                # Add a "double check" section.
                # This is used to ensure that the loop does not prematurely exit when the last file tests still
                # progressing that have not yet made it to the "files_to_process" queue.
                double_check += 1
                if double_check > 5:
                    # There are not more files to test. Mark manager threads as completed
                    self.stop_all_file_test_managers()
                    break
                self.event.wait(1)
                continue

            # Calculate percent of files tested
            if not self.files_to_test.empty():
                current_queue_size = self.files_to_test.qsize()
                if int(total_file_count) > 0 and int(current_queue_size) > 0:
                    percent_remaining = int((int(current_queue_size) / int(total_file_count)) * 100)
                    percent_completed = int(100 - percent_remaining)
                    percent_completed_string = '{}% - Testing: {}'.format(percent_completed, current_file)
                elif current_file:
                    percent_completed_string = '{}% - Testing: {}'.format('???', current_file)

            # Fetch frontend messages from queue
            if not status_updates.empty():
                current_file = status_updates.get()
                continue
            elif not self.files_to_process.empty():
                item = self.files_to_process.get()
                self.add_path_to_queue(item.get('path'), library_id, item.get('priority_score'))
                continue
            else:
                self.event.wait(.1)

        # Wait for threads to finish
        for manager_id in self.file_test_managers:
            self.file_test_managers[manager_id].abort_flag.set()
            self.file_test_managers[manager_id].join(2)

        self._log("Library scan completed in {} seconds".format((time.time() - start_time)), level="warning")

        # Run a manual garbage collection
        gc.collect()

        # Remove frontend status message
        frontend_messages.remove_item('libraryScanProgress')

    def register_unmanic(self):
        from unmanic.libs import session
        s = session.Session()
        s.register_unmanic()
